use std::{iter::Peekable, num::ParseIntError, path::Path, str::Chars, sync::Arc};

// TODO: tree_sitter ?

#[derive(Debug, PartialEq)]
enum BareToken {
    Integer(i64),
    Plus,
    Minus,
    Star,
    Slash,
    Percent,
    OpenParen,
    CloseParen,
    NewLine,
    EndOfFile,
    // TODO: how to handle indentation?
}

#[derive(Debug, PartialEq)]
pub struct Token {
    location: Option<Location>, // Not all tokens are associated with a location, such as EndOfFile
    token: BareToken,
}

#[derive(Debug, PartialEq, Clone, Default)]
pub enum Source {
    #[default]
    Unknown,

    File(Arc<Box<Path>>),

    Repl(usize),
}

#[derive(Debug, PartialEq, Clone)]
struct Location {
    source: Source, // TODO: not certain how this will work in the real world, but I don't want all these locations taking up a bunch of memory unnecessarily as we lex and parse, so for now we deal with lifetimes so we can store references and have Source be cheap to clone everywhere
    line: usize,
    len: usize,
    col: usize,
}

#[derive(Debug, PartialEq)]
enum LexerError {
    // TODO: not a number error?
    ParseIntError(ParseIntError),
    Unexpected(char),
}

impl From<ParseIntError> for LexerError {
    fn from(value: ParseIntError) -> Self {
        LexerError::ParseIntError(value)
    }
}

#[derive(Debug, PartialEq)]
pub struct Error {
    pub location: Location,
    pub error: LexerError,
}

pub struct Lexer<'a> {
    line: usize,
    col: usize,
    source: Source,
    chars: Peekable<Chars<'a>>,
    collected: Vec<char>,

    done: bool,
    sent_eof: bool,
}

impl<'a> Lexer<'a> {
    pub fn new(code: &'a str, source: Source) -> Self {
        let lexer = Lexer {
            done: false,
            sent_eof: false,
            line: 1,
            col: 0,
            source: source.clone(),
            chars: code.chars().peekable(),
            collected: vec![],
        };
        lexer
    }

    fn advance(&mut self) {
        if self.done {
            return;
        }
        match self.chars.next() {
            Some('\n') => {
                self.line += 1;
                self.col = 1;
                self.collected.push('\n')
            }
            Some(c) => {
                self.col += 1;
                self.collected.push(c)
            }
            None => self.done = true,
        }
    }

    fn current_location(&self) -> Location {
        Location {
            source: self.source.clone(),
            line: self.line,
            len: self.collected.len(),
            col: self.col,
        }
    }

    fn produce(&mut self, token: BareToken) -> Token {
        let t = Token {
            location: Some(self.current_location()),
            token,
        };
        self.collected.clear();
        t
    }

    fn produce_error(&self, error: LexerError) -> Error {
        Error {
            location: self.current_location(),
            error,
        }
    }

    fn token(&mut self) -> Result<Option<Token>, Error> {
        self.advance();
        let current = match self.collected.last() {
            None => return Ok(None),
            Some(c) => c,
        };
        Ok(Some(match current {
            c if c.is_digit(10) => self.number()?,
            '\n' => self.produce(BareToken::NewLine),
            '+' => self.produce(BareToken::Plus),

            c if c.is_whitespace() => {
                self.whitespace();
                return self.token();
            }

            c => return Err(self.produce_error(LexerError::Unexpected(*c))),
        }))
    }

    fn whitespace(&mut self) {
        while let Some(c) = self.chars.peek() {
            if c.is_whitespace() {
                self.advance()
            } else {
                break;
            }
        }
        self.collected.clear();
    }

    fn number(&mut self) -> Result<Token, Error> {
        while let Some(c) = self.chars.peek() {
            if c.is_ascii_digit() {
                self.advance()
            } else {
                break;
            }
        }

        Ok(self.produce(BareToken::Integer(
            self.collected
                .iter()
                .collect::<String>()
                .parse()
                .map_err(|e: ParseIntError| self.produce_error(e.into()))?,
        )))
    }
}

impl<'a> Iterator for Lexer<'a> {
    type Item = Result<Token, Error>;

    fn next(&mut self) -> Option<Self::Item> {
        if self.done && self.sent_eof {
            return None;
        }
        if self.done && !self.sent_eof {
            self.sent_eof = true;
            return Some(Ok(Token {
                location: None,
                token: BareToken::EndOfFile,
            }));
        }
        match self.token() {
            Ok(Some(t)) => return Some(Ok(t)),
            Ok(None) => return self.next(),
            Err(e) => return Some(Err(e)),
        }
    }
}

// pub fn lex_str(code: &str) -> Result<Vec<Token>, Error> {
//     lex(code, None)
// }

// pub fn lex(code: &str, source: Option<Source>) -> Result<Vec<Token>, Error> {
//     let source: Source = source.unwrap_or_default();
//     let mut result: Vec<Token> = vec![];
//     let mut scanner = scanner(source.clone(), code).peekable();
//     loop {
//         match Self::token(&mut scanner) {
//             Ok(Some(t)) => {
//                 let end = t.token == BareToken::EndOfFile;
//                 result.push(t);
//                 if end {
//                     break;
//                 }
//             }
//             Ok(None) => {}
//             Err(e) => return Err(e),
//         }
//     }
//     return Ok(result);
// }

// fn token() -> Result<Option<Token>, Error> {
//     let r = match scanner.next() {
//         None => {
//             return Ok(Some(Token {
//                 token: BareToken::EndOfFile,
//                 location: None,
//             }))
//         }
//         Some(s) => s,
//     };
//     match r.c {
//         c if c.is_ascii_whitespace() => Ok(None),
//         '\n' => t(BareToken::NewLine),
//         '(' => t(BareToken::OpenParen),
//         ')' => t(BareToken::CloseParen),
//         '0'..='9' => Self::number(r.c, r.location, scanner),
//         '+' => t(BareToken::Plus),
//         '-' => t(BareToken::Minus),
//         '*' => t(BareToken::Star),
//         '/' => t(BareToken::Slash),
//         '%' => t(BareToken::Percent),
//         c => {
//             return Err(Error {
//                 location: r.location,
//                 error: LexerError::Unexpected(c),
//             })
//         }
//     }
// }

// fn number(
//     first_digit: char,
//     mut location: Location,
//     scanner: &mut impl Iterator<Item = ScannerEntry>,
// ) -> Result<Token, Error> {
//     let mut scanner = scanner.peekable();
//     let s = iter::once(first_digit)
//         .chain(from_fn(move || {
//             scanner
//                 .by_ref()
//                 .next_if(|r| r.c.is_ascii_digit())
//                 .map(|r| r.c)
//         }))
//         .collect::<String>();
//     location.len += s.len();
//     let result = match s.parse() {
//         Ok(i) => BareToken::Integer(i),
//         Err(e) => {
//             return Err(Error {
//                 error: LexerError::ParseIntError(e),
//                 location,
//             })
//         }
//     };
//     Ok()
// }

#[cfg(test)]
mod test {
    use super::*;

    fn pos(col: usize) -> Option<Location> {
        Some(Location {
            line: 1,
            len: 1,
            col,
            source: Source::Unknown,
        })
    }

    fn t(location: Option<Location>, token: BareToken) -> Token {
        Token { location, token }
    }

    #[test]
    fn addition_operation() -> Result<(), Error> {
        let lexer = Lexer::new("3 + 9", Source::Unknown);
        let tokens: Result<Vec<Token>, Error> = lexer.collect();
        assert_eq!(
            tokens,
            Ok(vec![
                t(pos(1), BareToken::Integer(3)),
                t(pos(3), BareToken::Plus),
                t(pos(5), BareToken::Integer(9)),
                t(Option::None, BareToken::EndOfFile),
            ]),
        );
        assert_eq!(tokens?.len(), 4);
        Ok(())
    }
}
